from ..base import ChineseName as _GenericChineseName
from ..base import EnglishName as _GenericEnglishName


class ChineseName(_GenericChineseName):
    pass


class EnglishName(_GenericEnglishName):
    pass
