_GAMES = {
    'arknights': ('アークナイツ', 'アークナイツ'),
    'fgo': ('Fate/GrandOrder', 'Fate'),
    'azurlane': ('アズールレーン', 'アズールレーン'),
    'genshin': ('原神', '原神'),
    'girlsfrontline': ('ドールズフロントライン', 'ドールズフロントライン'),
    'neuralcloud': ('云图计划', '云图计划'),
    'bluearchive': ('ブルーアーカイブ', 'ブルーアーカイブ'),
    'pathtonowhere': ('無期迷途', '無期迷途'),
    'nikke': ('勝利の女神:NIKKE', '勝利の女神:NIKKE'),
    'starrail': ('崩壊スターレイル', 'スターレイル'),
}
